const coral = @import("coral");

const ext = @import("./ext.zig");

pub const file = @import("./file.zig");

const heap = @import("./heap.zig");

const kym = @import("./kym.zig");

const App = struct {
	renderer: *ext.SDL_Renderer,
	window: *ext.SDL_Window,
	systems: Systems,
	target_frame_time: f64,

	keys: struct {
		pressed: [ext.SDL_NUM_SCANCODES]bool = [_]bool{false} ** ext.SDL_NUM_SCANCODES,
		released: [ext.SDL_NUM_SCANCODES]bool = [_]bool{false} ** ext.SDL_NUM_SCANCODES,
		held: [ext.SDL_NUM_SCANCODES]bool = [_]bool{false} ** ext.SDL_NUM_SCANCODES,
	},

	const DynamicObject = struct {
		self: *App,

		fn get(context: kym.Typeinfo.GetContext) kym.RuntimeError!?*kym.RuntimeObj {
			const index = context.get_index();

			defer context.env.release(index);

			const on_update_symbol = (try context.env.new_symbol("on_update")).pop().?;

			defer context.env.release(on_update_symbol);

			if (index.equals(on_update_symbol)) {
				return (try context.env.new_syscall(on_update)).pop();
			}

			return null;
		}

		fn on_update(env: *kym.RuntimeEnv) kym.RuntimeError!?*kym.RuntimeObj {
			const caller_object = try env.expect_object(try env.get_caller());

			defer env.release(caller_object);

			const app = @as(*DynamicObject, @ptrCast(@alignCast(try env.expect_dynamic(caller_object, typeinfo))));

			const callable = try env.expect_object((try env.arg(0)).pop());

			errdefer env.release(callable);

			try app.self.systems.push_one(callable);

			return null;
		}

		const typeinfo = &kym.Typeinfo{
			.name = "ona",
			.size = @sizeOf(DynamicObject),
			.get = get,
		};
	};

	pub const InitError = error {
		RendererUnavailable,
		WindowUnavailable,
	};

	const Systems = coral.list.Stack(*kym.RuntimeObj);

	pub fn as_importer(self: *App) kym.Importer {
		return kym.Importer.bind(App, self, import);
	}

	pub fn deinit(self: *App) void {
		ext.SDL_DestroyWindow(self.window);
		ext.SDL_DestroyRenderer(self.renderer);
		self.systems.deinit();
	}

	fn import(self: *App, env: *kym.RuntimeEnv) kym.RuntimeError!?*kym.RuntimeObj {
		return (try env.new_dynamic(DynamicObject.typeinfo, DynamicObject{.self = self})).pop();
	}

	pub fn init() InitError!App {
		const window = create: {
			const height = 480;
			const width = 640;
			const pos = ext.SDL_WINDOWPOS_CENTERED;
			const flags = ext.SDL_WINDOW_HIDDEN;

			break: create ext.SDL_CreateWindow("", pos, pos, width, height, flags) orelse {
				return error.WindowUnavailable;
			};
		};

		errdefer ext.SDL_DestroyWindow(window);

		const renderer = create: {
			const default_driver_index = -1;
			const flags = ext.SDL_RENDERER_ACCELERATED;

			break: create ext.SDL_CreateRenderer(window, default_driver_index, flags) orelse {
				return error.RendererUnavailable;
			};
		};

		errdefer ext.SDL_DestroyRenderer(renderer);

		return .{
			.systems = Systems.init(heap.allocator),
			.target_frame_time = 1.0 / 60.0,
			.renderer = renderer,
			.window = window,
			.keys = .{},
		};
	}

	pub fn load(self: *App, env: *kym.RuntimeEnv) kym.RuntimeError!void {
		var previous_width = @as(c_int, 0);
		var previous_height = @as(c_int, 0);

		ext.SDL_GetWindowSize(self.window, &previous_width, &previous_height);

		var width = previous_width;
		var height = previous_height;

		defer {
			const pos = ext.SDL_WINDOWPOS_CENTERED;

			ext.SDL_SetWindowSize(self.window, width, height);
			ext.SDL_SetWindowPosition(self.window, pos, pos);
			ext.SDL_ShowWindow(self.window);
		}

		if ((try env.import(file.Path.of(&.{"app.ona"}))).pop()) |manifest| {
			defer env.release(manifest);

			if (try kym.get_field(env, manifest, "width")) |object| {
				defer env.release(object);

				const fixed = try env.expect_fixed(object);

				if (fixed > 0 and fixed < coral.math.max_int(@typeInfo(@TypeOf(width)).Int)) {
					width = @intCast(fixed);
				}
			}

			if (try kym.get_field(env, manifest, "height")) |object| {
				defer env.release(object);

				const fixed = try env.expect_fixed(object);

				if (fixed > 0 and fixed < coral.math.max_int(@typeInfo(@TypeOf(height)).Int)) {
					height = @intCast(fixed);
				}
			}

			if (try kym.get_field(env, manifest, "tick_rate")) |object| {
				defer env.release(object);

				self.target_frame_time = 1.0 / try env.expect_float(object);
			}

			if (try kym.get_field(env, manifest, "title")) |object| {
				defer env.release(object);

				const title = coral.utf8.SmallString.from_units(try env.expect_string(object)) catch |from_error| {
					return switch (from_error) {
						error.InvalidUtf8 => env.raise(error.TypeMismatch, "`title` cannot contain invalid utf8", .{}),
						error.TooBig => env.raise(error.TypeMismatch, "`title` is too long", .{}),
					};
				};

				ext.SDL_SetWindowTitle(self.window, title.to_units());
			} else {
				ext.SDL_SetWindowTitle(self.window, "Ona");
			}
		}
	}

	pub fn poll(self: *App) bool {
		var event = @as(ext.SDL_Event, undefined);

		coral.io.zero(@ptrCast(&self.keys.pressed));
		coral.io.zero(@ptrCast(&self.keys.released));

		while (ext.SDL_PollEvent(&event) != 0) {
			switch (event.type) {
				ext.SDL_QUIT => return false,

				ext.SDL_KEYDOWN => {
					self.keys.pressed[event.key.keysym.scancode] = true;
					self.keys.held[event.key.keysym.scancode] = true;
				},

				ext.SDL_KEYUP => {
					self.keys.released[event.key.keysym.scancode] = true;
					self.keys.held[event.key.keysym.scancode] = false;
				},

				else => {},
			}
		}

		return true;
	}

	pub fn render(self: *App) void {
		_ = ext.SDL_SetRenderDrawColor(self.renderer, 0, 0, 0, 255);
		_ = ext.SDL_RenderClear(self.renderer);
		_ = ext.SDL_RenderPresent(self.renderer);
	}

	pub fn update(self: *App, env: *kym.RuntimeEnv, time_delta: f32) kym.RuntimeError!void {
		for (self.systems.values) |system| {
			(try (try (try env.new_float(time_delta)).push(system)).call(1, null)).discard();
		}
	}
};

fn last_sdl_error() [:0]const u8 {
	return coral.io.slice_sentineled(@as(u8, 0), @as([*:0]const u8, @ptrCast(ext.SDL_GetError())));
}

fn load_prompt() void {

}

pub fn log_info(message: []const coral.io.Byte) void {
	ext.SDL_LogInfo(
		ext.SDL_LOG_CATEGORY_APPLICATION,
		"%.*s",
		coral.math.clamped_cast(@typeInfo(c_int).Int, message.len),
		message.ptr,
	);
}

pub fn log_warn(message: []const coral.io.Byte) void {
	ext.SDL_LogWarn(
		ext.SDL_LOG_CATEGORY_APPLICATION,
		"%.*s",
		coral.math.clamped_cast(@typeInfo(c_int).Int, message.len),
		message.ptr,
	);
}

pub fn log_fail(message: []const coral.io.Byte) void {
	ext.SDL_LogError(
		ext.SDL_LOG_CATEGORY_APPLICATION,
		"%.*s",
		coral.math.clamped_cast(@typeInfo(c_int).Int, message.len),
		message.ptr,
	);
}

const milliseconds_per_second = 1000;

pub fn run_app(file_access: file.Access) void {
	defer heap.trace_leaks();

	if (ext.SDL_Init(ext.SDL_INIT_EVERYTHING) != 0) {
		return log_fail(last_sdl_error());
	}

	defer ext.SDL_Quit();

	var app = App.init() catch |init_error| {
		return log_fail(switch (init_error) {
			error.WindowUnavailable => "unable to acquire window",
			error.RendererUnavailable => "unable to acquire renderer",
		});
	};

	defer app.deinit();

	var script_env = kym.RuntimeEnv.init(heap.allocator, 255, .{
		.import_access = file_access,
		.print_out = log_info,
		.print_err = log_fail,
	}) catch {
		return log_fail("failed to initialize script runtime");
	};

	defer script_env.deinit();

	script_env.override_import(file.Path.of(&.{"ona"}), app.as_importer()) catch return;

	app.load(&script_env) catch load_prompt();

	var ticks_previous = ext.SDL_GetTicks64();
	var accumulated_time = @as(f64, 0);

	running_loop: while (true) {
		const ticks_current = ext.SDL_GetTicks64();
		const delta_time = @as(f64, @floatFromInt(ticks_current - ticks_previous)) / milliseconds_per_second;

		ticks_previous = ticks_current;
		accumulated_time += delta_time;

		while (accumulated_time >= app.target_frame_time) : (accumulated_time -= app.target_frame_time) {
			if (!app.poll()) {
				break: running_loop;
			}

			app.update(&script_env, @floatCast(app.target_frame_time)) catch return;
		}

		app.render();
		ext.SDL_Delay(1);
	}
}
